use nom::branch::alt;
use nom::bytes::complete::is_not;
use nom::bytes::complete::tag;
use nom::character::complete::line_ending;
use nom::character::complete::space0;
use nom::combinator::eof;
use nom::combinator::not;
use nom::combinator::peek;
use nom::combinator::recognize;
use nom::combinator::verify;
use nom::multi::many1;
use nom::multi::many_till;
use nom::sequence::tuple;

use super::Context;
use crate::error::Res;
use crate::parser::exiting::ExitClass;
use crate::parser::greater_element::TableRow;
use crate::parser::lesser_element::TableCell;
use crate::parser::object::Object;
use crate::parser::object_parser::minimal_set_object;
use crate::parser::parser_context::ContextElement;
use crate::parser::parser_context::ExitMatcherNode;
use crate::parser::parser_with_context::parser_with_context;
use crate::parser::util::exit_matcher_parser;
use crate::parser::util::get_consumed;
use crate::parser::util::start_of_line;
use crate::parser::Table;

/// Parse an org-mode-style table
///
/// This is not the table.el style.
#[tracing::instrument(ret, level = "debug")]
pub fn org_mode_table<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, Table<'s>> {
    start_of_line(context, input)?;
    peek(tuple((space0, tag("|"))))(input)?;

    let parser_context = context
        .with_additional_node(ContextElement::ConsumeTrailingWhitespace(true))
        .with_additional_node(ContextElement::Context("table"))
        .with_additional_node(ContextElement::ExitMatcherNode(ExitMatcherNode {
            class: ExitClass::Alpha,
            exit_matcher: &table_end,
        }));

    let org_mode_table_row_matcher = parser_with_context!(org_mode_table_row)(&parser_context);
    let exit_matcher = parser_with_context!(exit_matcher_parser)(&parser_context);

    let (remaining, (children, _exit_contents)) =
        many_till(org_mode_table_row_matcher, exit_matcher)(input)?;

    // TODO: Consume trailing formulas
    let source = get_consumed(input, remaining);

    Ok((remaining, Table { source, children }))
}

#[tracing::instrument(ret, level = "debug")]
fn table_end<'r, 's>(context: Context<'r, 's>, input: &'s str) -> Res<&'s str, &'s str> {
    start_of_line(context, input)?;
    recognize(tuple((space0, not(tag("|")))))(input)
}

#[tracing::instrument(ret, level = "debug")]
pub fn org_mode_table_row<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, TableRow<'s>> {
    alt((
        parser_with_context!(org_mode_table_row_rule)(context),
        parser_with_context!(org_mode_table_row_regular)(context),
    ))(input)
}

#[tracing::instrument(ret, level = "debug")]
pub fn org_mode_table_row_rule<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, TableRow<'s>> {
    start_of_line(context, input)?;
    let (remaining, _) = tuple((space0, tag("|-"), is_not("\r\n"), line_ending))(input)?;
    let source = get_consumed(input, remaining);
    Ok((
        remaining,
        TableRow {
            source,
            children: Vec::new(),
        },
    ))
}

#[tracing::instrument(ret, level = "debug")]
pub fn org_mode_table_row_regular<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, TableRow<'s>> {
    start_of_line(context, input)?;
    let (remaining, _) = tuple((space0, tag("|")))(input)?;
    let (remaining, children) =
        many1(parser_with_context!(org_mode_table_cell)(context))(remaining)?;
    let (remaining, _tail) = recognize(tuple((space0, alt((line_ending, eof)))))(remaining)?;
    let source = get_consumed(input, remaining);
    Ok((remaining, TableRow { source, children }))
}

#[tracing::instrument(ret, level = "debug")]
pub fn org_mode_table_cell<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, TableCell<'s>> {
    let parser_context =
        context.with_additional_node(ContextElement::ExitMatcherNode(ExitMatcherNode {
            class: ExitClass::Beta,
            exit_matcher: &org_mode_table_cell_end,
        }));
    let table_cell_set_object_matcher =
        parser_with_context!(table_cell_set_object)(&parser_context);
    let exit_matcher = parser_with_context!(exit_matcher_parser)(&parser_context);
    let (remaining, (children, _exit_contents)) = verify(
        many_till(table_cell_set_object_matcher, exit_matcher),
        |(children, exit_contents)| !children.is_empty() || exit_contents.ends_with("|"),
    )(input)?;

    let (remaining, _tail) = org_mode_table_cell_end(&parser_context, remaining)?;

    let source = get_consumed(input, remaining);

    Ok((remaining, TableCell { source, children }))
}

#[tracing::instrument(ret, level = "debug")]
fn org_mode_table_cell_end<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, &'s str> {
    recognize(tuple((space0, alt((tag("|"), peek(line_ending))))))(input)
}

#[tracing::instrument(ret, level = "debug")]
pub fn table_cell_set_object<'r, 's>(
    context: Context<'r, 's>,
    input: &'s str,
) -> Res<&'s str, Object<'s>> {
    not(|i| context.check_exit_matcher(i))(input)?;

    parser_with_context!(minimal_set_object)(context)(input)
    // TODO: add citations, export snippets, footnote references, links, macros, radio targets, targets, and timestamps.
}
